/*
Copyright 2019 ZEYEN Olivier, PÉCHOUX Romain, HAINRY Emmanuel, JEANDEL Emmanuel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */
package complexityparser.types.env;

import complexityparser.types.Tier;
import complexityparser.types.FOTier;
import complexityparser.types.FOTierList;
import complexityparser.tos.TOS;
import lib.Mask;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CharStreams;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTreeWalker;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * A singleton class encoding typing environments. A hashmap relating an
 * operation (operator, method, ...) namespace to the corresponding operator.
 * (An operator has an attribute results of type ArrayList<FOTier> containing
 * an arraylist of admissible FOTier).
 */
public class TypingEnv {

    private static TypingEnv instance = null;
    private HashMap<String, FOTierList> env;

    protected TypingEnv() {
        env = new HashMap<>();
    }

    /**
     * Initializes the singleton using the file 'operators.config' in the
     * working directory.
     *
     * @throws IOException
     */
    public static void loadInstanceFromFile() throws IOException {
        instance = loadFromFile("operators.config");
        if (instance == null) {
            throw new IOException("operators.config load failed");
        }
    }

    /**
     * Removes any entry of the hashmap env, starting from <root>. The entries
     * are supposed to be method entries generated by the visitor parser.
     */
    public void clear() {
        ArrayList<String> keys = new ArrayList<>();
        for (Map.Entry<String, FOTierList> entry : env.entrySet()) {
            if (entry.getKey().contains("<root>")) {
                keys.add(entry.getKey());
            }
        }
        for (String k : keys) {
            env.remove(k);
        }
    }

    /**
     *
     * @return the attribute instance (null if the method
     * #loadInstanceFromFIle() has not been called yet).
     */
    public static TypingEnv getInstance() {
        return instance;
    }

    /**
     * Adds a FOTierList and a name to the typing environment.
     *
     * @param k - a name.
     * @param o - an operator.
     */
    public void add(String k, FOTierList o) {
        env.put(k, o);
    }

    /**
     *
     * @param name - a namespace.
     * @param in - input tiers.
     * @return a FOTier matching a namespace and input tiers given as
     * parameters. A default FOTier if no match is found.
     */
    public FOTier findRes(String name, Tier... in) {
        if (env.containsKey(name)) {
            return env.get(name).findRes(in);
        }
        return new FOTier(Tier.None, Tier.None, new Tier[0]);
    }

    /**
     *
     * @param name - a name.
     * @param m - a mask.
     * @param in - input tiers.
     * @return a FOTier matching a namespace, a mask, and input tiers. A default
     * FOTier if no match is found.
     */
    public FOTier findCompatibleIn(String name, Mask m, Tier... in) {
        if (env.containsKey(name)) {
            return env.get(name).findCompatibleRes(m, in);
        }
        return new FOTier(Tier.None, Tier.None, new Tier[0]);
    }

    /**
     * The behaviour is different from #findRes(String name, Tier... in):
     * Additional checks are performed on overriden methods.
     *
     * @param name - a namespace.
     * @param in - input tiers.
     * @return the FOTier matching a method namespace and input tiers.
     */
    public FOTier findMethod(String name, Tier... in) {
        FOTier res = null;
        TOS tos = TOS.getInstance();
        while (!name.contains("<root>::Object::") && !env.containsKey(name)) {
            name = tos.getParentNamespace(name);
        }
        //If the namespace is of the form "<root>::Object::..." then the method has not been found in the class hierarchy.
        if (!name.contains("<root>::Object::") && env.containsKey(name)) {
            res = env.get(name).findRes(in);
            res = checkOverrideMethod(name, res, in);
        }
        return res;
    }

    /**
     * The behaviour is different from {@link TypingEnv#findCompatibleIn(String name, Mask m,
     * Tier... in)}: Additional checks are performed on overriden methods.
     *
     * @param name - a name.
     * @param m - a mask.
     * @param in - input tiers.
     * @return a compatible FOTier matching a method namespace, mask, and input
     * tiers.
     */
    public FOTier findCompatibleMethod(String name, Mask m, Tier... in) {
        TOS tos = TOS.getInstance();
        while (!name.contains("<root>::Object::") && !env.containsKey(name)) {
            name = tos.getParentNamespace(name);
        }
        if (!name.contains("<root>::Object") && env.containsKey(name)) {
            FOTier[] res = env.get(name).findAllCompatibles(m, in);
            for (FOTier i : res) {
                FOTier tmp = checkOverrideMethod(name, i, i.getIn());
                if (!tmp.equals(new FOTier(Tier.None, Tier.None, new Tier[0]))) {
                    return tmp;
                }
            }
        }
        return new FOTier(Tier.None, Tier.None, new Tier[0]);
    }

    /**
     * Methods returning a primitive type do not need to return exactly the same
     * tier. Methods returning reference type need to preserve exactly the tier.
     * The check will be performed on every subclass to handle method overrides.
     *
     * @param name - a namespace.
     * @param out - a FOTier.
     * @param in - input tiers.
     * @return a FOTier matching the input parameters. A default FOTier if there
     * is no match.
     */
    public FOTier checkOverrideMethod(String name, FOTier out, Tier... in) {
        TOS tos = TOS.getInstance();
        //Only the children need to be verified since no superclass can be used in case of polymorphism.
        return checkOverrideMethod_aux(name, out, in);
    }

    /**
     * A recursive algorithm to check the if the method of namespace, input
     * tiers, and FOTier are compatible with the class hierarchy.
     *
     * @param name - a namespace.
     * @param out - a FOTier.
     * @param in - input tiers.
     * @return a FOTier matching the inputs. A default FOTier if there is no
     * match.
     */
    private FOTier checkOverrideMethod_aux(String name, FOTier out, Tier... in) {
        FOTier res = new FOTier(out.getOut(), out.getEnv(), in);
        if (env.containsKey(name)) {
            FOTier tmp = env.get(name).findRes(in);
            /*
            The method return type has to be typed by considering possible method overrides:
            --if the return type is a primitive type: the output tier is the minimum of input tiers;
            --else, if the return type is a reference type, all the tiers must be equal.
            The environment tier is the maximum of the input tiers.
             */
            res.setEnv(Tier.max(tmp.getEnv(), out.getEnv()));
            if (env.get(name).isPrimitiveReturnType()) {
                res.setOut(Tier.min(tmp.getOut(), out.getOut()));
            } else {
                if (res.getOut() != out.getOut()) {
                    return new FOTier(Tier.None, Tier.None, new Tier[0]);
                }
            }
        }
        TOS tos = TOS.getInstance();
        String[] children = tos.getChildrenNamespaces(name);
        if (children != null) {
            for (int i = 0; i < children.length; i++) {
                FOTier tmp = checkOverrideMethod_aux(children[i], res, in);
                res.setEnv(Tier.max(tmp.getEnv(), out.getEnv()));
                if (env.get(name).isPrimitiveReturnType()) {
                    res.setOut(Tier.min(tmp.getOut(), out.getOut()));
                } else {
                    if (res.getOut() != out.getOut()) {
                        return new FOTier(Tier.None, Tier.None, new Tier[0]);
                    }
                }
            }
        }
        return res;
    }

    /**
     * Loads the operators FOTiers from a file using the ANTLR grammar and the
     * OperatorTypingEnvBuilder class.
     *
     * @param fileName - a file name.
     * @return the singleton instance of TypingEnv.
     */
    private static TypingEnv loadFromFile(String fileName) {
        TypingEnv res = null;
        byte[] input = null;
        try {
            Path p = new File(fileName).toPath();
            input = Files.readAllBytes(p);
        } catch (IOException e) {
            e.printStackTrace();
            return res;
        }
        String txt = new String(input, StandardCharsets.UTF_8);
        CharStream stream = CharStreams.fromString(txt);
        OperatorsLexer lexer = new OperatorsLexer(stream);
        CommonTokenStream token = new CommonTokenStream(lexer);
        OperatorsParser parser = new OperatorsParser(token);
        OperatorsParser.CompilationUnitContext context = parser.compilationUnit();
        OperatorTypingEnvBuilder listener = new OperatorTypingEnvBuilder();
        ParseTreeWalker.DEFAULT.walk(listener, context);
        res = listener.getOp();
        return res;
    }

    /**
     *
     * @param namespace - a namespace.
     * @return the operator associated with the given namespace.
     */
    public FOTierList get(String namespace) {
        return env.get(namespace);
    }

    @Override
    public String toString() {
        StringBuilder str = new StringBuilder();
        for (Map.Entry<String, FOTierList> e : env.entrySet()) {
            str.append("\"");
            str.append(e.getKey());
            str.append("\" {\n");
            str.append(e.getValue());
            str.append("}\n");
        }
        return str.toString();
    }
}
